@@ -50,11 +50,7 @@
 struct _MateDesktopThumbnailFactoryPrivate {
   MateDesktopThumbnailSize size;
 
-#if GLIB_CHECK_VERSION(2, 31, 0)
   GMutex lock;
-#else
-  GMutex* lock;
-#endif
 
   GList *thumbnailers;
   GHashTable *mime_types_map;
@@ -307,13 +303,11 @@ create_loader (GFile        *file,
                const guchar *data,
                gsize         size)
 {
-  GdkPixbufLoader *loader;
+  GdkPixbufLoader *loader = NULL;
   GError *error = NULL;
   char *mime_type;
   char *filename;
 
-  loader = NULL;
-
   /* need to specify the type here because the gdk_pixbuf_loader_write
      doesn't have access to the filename in order to correct detect
      the image type. */
@@ -325,7 +319,7 @@ create_loader (GFile        *file,
     loader = gdk_pixbuf_loader_new_with_mime_type (mime_type, &error);
   }
 
-  if (loader == NULL) {
+  if (loader == NULL && error != NULL) {
     g_warning ("Unable to create loader for mime type %s: %s", mime_type, error->message);
     g_clear_error (&error);
     loader = gdk_pixbuf_loader_new ();
@@ -499,15 +493,7 @@ mate_desktop_thumbnail_factory_finalize (GObject *object)
       priv->monitors = NULL;
     }
 
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_clear (&priv->lock);
-#else
-  if (priv->lock)
-    {
-      g_mutex_free (priv->lock);
-      priv->lock = NULL;
-    }
-#endif
 
   if (priv->disabled_types)
     {
@@ -592,11 +578,7 @@ update_or_create_thumbnailer (MateDesktopThumbnailFactory *factory,
   Thumbnailer *thumb;
   gboolean found = FALSE;
 
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_lock (&priv->lock);
-#else
-  g_mutex_lock (priv->lock);
-#endif
 
   for (l = priv->thumbnailers; l && !found; l = g_list_next (l))
     {
@@ -624,11 +606,7 @@ update_or_create_thumbnailer (MateDesktopThumbnailFactory *factory,
         mate_desktop_thumbnail_factory_add_thumbnailer (factory, thumb);
     }
 
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_unlock (&priv->lock);
-#else
-  g_mutex_unlock (priv->lock);
-#endif
 }
 
 static void
@@ -639,11 +617,7 @@ remove_thumbnailer (MateDesktopThumbnailFactory *factory,
   GList *l;
   Thumbnailer *thumb;
 
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_lock (&priv->lock);
-#else
-  g_mutex_lock (priv->lock);
-#endif
 
   for (l = priv->thumbnailers; l; l = g_list_next (l))
     {
@@ -661,11 +635,7 @@ remove_thumbnailer (MateDesktopThumbnailFactory *factory,
         }
     }
 
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_unlock (&priv->lock);
-#else
-  g_mutex_unlock (priv->lock);
-#endif
 }
 
 static void
@@ -767,11 +737,7 @@ external_thumbnailers_disabled_all_changed_cb (GSettings                    *set
 {
   MateDesktopThumbnailFactoryPrivate *priv = factory->priv;
 
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_lock (&priv->lock);
-#else
-  g_mutex_lock (priv->lock);
-#endif
 
   priv->disabled = g_settings_get_boolean (priv->settings, "disable-all");
   if (priv->disabled)
@@ -785,11 +751,7 @@ external_thumbnailers_disabled_all_changed_cb (GSettings                    *set
       mate_desktop_thumbnail_factory_load_thumbnailers (factory);
     }
 
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_unlock (&priv->lock);
-#else
-  g_mutex_unlock (priv->lock);
-#endif
 }
 
 static void
@@ -799,22 +761,14 @@ external_thumbnailers_disabled_changed_cb (GSettings                    *setting
 {
   MateDesktopThumbnailFactoryPrivate *priv = factory->priv;
 
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_lock (&priv->lock);
-#else
-  g_mutex_lock (priv->lock);
-#endif
 
   if (priv->disabled)
     return;
   g_strfreev (priv->disabled_types);
   priv->disabled_types = g_settings_get_strv (priv->settings, "disable");
 
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_unlock (&priv->lock);
-#else
-  g_mutex_unlock (priv->lock);
-#endif
 }
 
 static void
@@ -833,11 +787,7 @@ mate_desktop_thumbnail_factory_init (MateDesktopThumbnailFactory *factory)
                                                 (GDestroyNotify)g_free,
                                                 (GDestroyNotify)thumbnailer_unref);
 
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_init (&priv->lock);
-#else
-  priv->lock = g_mutex_new ();
-#endif
 
   priv->settings = g_settings_new ("org.mate.thumbnailers");
 
@@ -903,7 +853,7 @@ mate_desktop_thumbnail_factory_new (MateDesktopThumbnailSize size)
  *
  * Usage of this function is threadsafe.
  *
- * Return value: The absolute path of the thumbnail, or %NULL if none exist.
+ * Return value: (transfer full): The absolute path of the thumbnail, or %NULL if none exist.
  *
  * Since: 2.2
  **/
@@ -932,13 +882,8 @@ mate_desktop_thumbnail_factory_lookup (MateDesktopThumbnailFactory *factory,
 
   file = g_strconcat (g_checksum_get_string (checksum), ".png", NULL);
   
-#if GLIB_CHECK_VERSION (2, 34, 0)
   path = g_build_filename (g_get_user_cache_dir (),
                            "thumbnails",
-#else
-  path = g_build_filename (g_get_home_dir (),
-                           ".thumbnails",
-#endif
                            (priv->size == MATE_DESKTOP_THUMBNAIL_SIZE_NORMAL)?"normal":"large",
                            file,
                            NULL);
@@ -998,13 +943,8 @@ mate_desktop_thumbnail_factory_has_valid_failed_thumbnail (MateDesktopThumbnailF
 
   file = g_strconcat (g_checksum_get_string (checksum), ".png", NULL);
 
-#if GLIB_CHECK_VERSION (2, 34, 0)
   path = g_build_filename (g_get_user_cache_dir (),
                            "thumbnails/fail",
-#else
-  path = g_build_filename (g_get_home_dir (),
-                           ".thumbnails/fail",
-#endif
                            appname,
                            file,
                            NULL);
@@ -1024,52 +964,74 @@ mate_desktop_thumbnail_factory_has_valid_failed_thumbnail (MateDesktopThumbnailF
   return res;
 }
 
+/* forbidden/buggy GdkPixbufFormat names */
+const char *forbidden[] = { "tga", "icns", "jpeg2000" };
+
+static gboolean
+type_is_forbidden (const gchar *name)
+{
+    gint i = 0;
+    for (i = 0; i < G_N_ELEMENTS (forbidden); i++) {
+        if (g_strcmp0 (forbidden[i], name) == 0) {
+            return TRUE;
+        }
+    }
+    return FALSE;
+}
+
 static gboolean
 mimetype_supported_by_gdk_pixbuf (const char *mime_type)
 {
-        guint i;
-        static gsize formats_hash = 0;
-        gchar *key;
-        gboolean result;
+    guint i;
+    static gsize formats_hash = 0;
+    gchar *key;
+    gboolean result;
 
-	if (g_once_init_enter (&formats_hash)) {
-                GSList *formats, *list;
-		GHashTable *hash;
+    if (g_once_init_enter (&formats_hash)) {
+        GSList *formats, *list;
+        GHashTable *hash;
 
-                hash = g_hash_table_new_full (g_str_hash,
-					      (GEqualFunc) g_content_type_equals,
-					      g_free, NULL);
+        hash = g_hash_table_new_full (g_str_hash,
+                                      (GEqualFunc) g_content_type_equals,
+                                      g_free, NULL);
 
-                formats = gdk_pixbuf_get_formats ();
-                list = formats;
+        formats = gdk_pixbuf_get_formats ();
+        list = formats;
 
-                while (list) {
-                        GdkPixbufFormat *format = list->data;
-                        gchar **mime_types;
+        while (list) {
+            GdkPixbufFormat *format = list->data;
+            gchar **mime_types;
 
-                        mime_types = gdk_pixbuf_format_get_mime_types (format);
+            if (type_is_forbidden (format->name)) {
+                gdk_pixbuf_format_set_disabled (format, TRUE);
+                list = list->next;
+                continue;
+            }
 
-                        for (i = 0; mime_types[i] != NULL; i++)
-                                g_hash_table_insert (hash,
-                                                     (gpointer) g_content_type_from_mime_type (mime_types[i]),
-                                                     GUINT_TO_POINTER (1));	
+            mime_types = gdk_pixbuf_format_get_mime_types (format);
 
-                        g_strfreev (mime_types);
-                        list = list->next;
-                }
-                g_slist_free (formats);
+            for (i = 0; mime_types[i] != NULL; i++)
+                g_hash_table_insert (hash,
+                                     (gpointer) g_content_type_from_mime_type (mime_types[i]),
+                                     GUINT_TO_POINTER (1));
 
-		g_once_init_leave (&formats_hash, (gsize) hash);
+            g_strfreev (mime_types);
+            list = list->next;
         }
 
-        key = g_content_type_from_mime_type (mime_type);
-        if (g_hash_table_lookup ((void*)formats_hash, key))
-                result = TRUE;
-        else
-                result = FALSE;
-        g_free (key);
+        g_slist_free (formats);
 
-        return result;
+        g_once_init_leave (&formats_hash, (gsize) hash);
+    }
+
+    key = g_content_type_from_mime_type (mime_type);
+    if (g_hash_table_lookup ((void*)formats_hash, key))
+            result = TRUE;
+    else
+            result = FALSE;
+    g_free (key);
+
+    return result;
 }
 
 /**
@@ -1106,11 +1068,7 @@ mate_desktop_thumbnail_factory_can_thumbnail (MateDesktopThumbnailFactory *facto
   if (!mime_type)
     return FALSE;
 
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_lock (&factory->priv->lock);
-#else
-  g_mutex_lock (factory->priv->lock);
-#endif
   if (!mate_desktop_thumbnail_factory_is_disabled (factory, mime_type))
     {
       Thumbnailer *thumb;
@@ -1118,11 +1076,7 @@ mate_desktop_thumbnail_factory_can_thumbnail (MateDesktopThumbnailFactory *facto
       thumb = g_hash_table_lookup (factory->priv->mime_types_map, mime_type);
       have_script = thumbnailer_try_exec (thumb);
     }
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_unlock (&factory->priv->lock);
-#else
-  g_mutex_unlock (factory->priv->lock);
-#endif
 
   if (have_script || mimetype_supported_by_gdk_pixbuf (mime_type))
     {
@@ -1245,11 +1199,7 @@ mate_desktop_thumbnail_factory_generate_thumbnail (MateDesktopThumbnailFactory *
   pixbuf = NULL;
 
   script = NULL;
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_lock (&factory->priv->lock);
-#else
-  g_mutex_lock (factory->priv->lock);
-#endif
   if (!mate_desktop_thumbnail_factory_is_disabled (factory, mime_type))
     {
       Thumbnailer *thumb;
@@ -1258,11 +1208,7 @@ mate_desktop_thumbnail_factory_generate_thumbnail (MateDesktopThumbnailFactory *
       if (thumb)
         script = g_strdup (thumb->command);
     }
-#if GLIB_CHECK_VERSION(2, 31, 0)
   g_mutex_unlock (&factory->priv->lock);
-#else
-  g_mutex_unlock (factory->priv->lock);
-#endif
   
   if (script)
     {
@@ -1362,13 +1308,8 @@ make_thumbnail_dirs (MateDesktopThumbnailFactory *factory)
 
   res = FALSE;
 
-#if GLIB_CHECK_VERSION (2, 34, 0)
   thumbnail_dir = g_build_filename (g_get_user_cache_dir (),
                                     "thumbnails",
-#else
-  thumbnail_dir = g_build_filename (g_get_home_dir (),
-                                    ".thumbnails",
-#endif
                                     NULL);
   if (!g_file_test (thumbnail_dir, G_FILE_TEST_IS_DIR))
     {
@@ -1401,13 +1342,8 @@ make_thumbnail_fail_dirs (MateDesktopThumbnailFactory *factory)
 
   res = FALSE;
 
-#if GLIB_CHECK_VERSION (2, 34, 0)
   thumbnail_dir = g_build_filename (g_get_user_cache_dir (),
                                     "thumbnails",
-#else
-  thumbnail_dir = g_build_filename (g_get_home_dir (),
-                                    ".thumbnails",
-#endif
                                     NULL);
   if (!g_file_test (thumbnail_dir, G_FILE_TEST_IS_DIR))
     {
@@ -1481,13 +1417,8 @@ mate_desktop_thumbnail_factory_save_thumbnail (MateDesktopThumbnailFactory *fact
 
   file = g_strconcat (g_checksum_get_string (checksum), ".png", NULL);
 
-#if GLIB_CHECK_VERSION (2, 34, 0)
   path = g_build_filename (g_get_user_cache_dir (),
                            "thumbnails",
-#else
-  path = g_build_filename (g_get_home_dir (),
-                           ".thumbnails",
-#endif
                            (priv->size == MATE_DESKTOP_THUMBNAIL_SIZE_NORMAL)?"normal":"large",
                            file,
                            NULL);
@@ -1594,13 +1525,8 @@ mate_desktop_thumbnail_factory_create_failed_thumbnail (MateDesktopThumbnailFact
 
   file = g_strconcat (g_checksum_get_string (checksum), ".png", NULL);
 
-#if GLIB_CHECK_VERSION (2, 34, 0)
   path = g_build_filename (g_get_user_cache_dir (),
                            "thumbnails/fail",
-#else
-  path = g_build_filename (g_get_home_dir (),
-                           ".thumbnails/fail",
-#endif
                            appname,
                            file,
                            NULL);
@@ -1690,13 +1616,8 @@ mate_desktop_thumbnail_path_for_uri (const char         *uri,
   file = g_strconcat (md5, ".png", NULL);
   g_free (md5);
   
-#if GLIB_CHECK_VERSION (2, 34, 0)
   path = g_build_filename (g_get_user_cache_dir (),
                            "thumbnails",
-#else
-  path = g_build_filename (g_get_home_dir (),
-                           ".thumbnails",
-#endif
                            (size == MATE_DESKTOP_THUMBNAIL_SIZE_NORMAL)?"normal":"large",
                            file,
                            NULL);
